/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include "marshalers.h"
#include "nntpgrab_glue.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#define NNTPGRAB_TYPE_CORE              (core_get_type ())
#define NNTPGRAB_CORE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_TYPE_CORE, NNTPGrabCore))
#define NNTPGRAB_CORE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))
#define IS_NNTPGRAB_CORE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_TYPE_CORE))
#define IS_NNTPGRAB_CORE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_TYPE_CORE))
#define NNTPGRAB_CORE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))

#ifdef WIN32
#define CLOSE(x)    closesocket(x)
#else
#define CLOSE(x)    close(x)
#endif

#ifdef _MSC_VER
#undef atoll
#define atoll _atoi64

#undef gai_strerror
#define gai_strerror gai_strerrorA
#endif

#define DEBUG(x)                            \
    g_print(x);                             \
    nntpgrab_core_emit_debug_message(x);

#define READ_TIMEOUT_VALUE      30

typedef struct _connection {
    int socket_id;
    char recv_buf[4096];      // It should be safe to assume that one line is less than 4096 characters
    int recv_buf_len;
} Connection;

struct NNTPGrabCore
{
    GObject parent;

    Connection socket_listener;
    Connection socket_command;
};

static NNTPGrabCore *core = NULL;

struct NNTPGrabCoreClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    PART_DOWNLOAD_START_SIGNAL,
    PART_DONE_SIGNAL,
    PART_FAILED_SIGNAL,
    PART_PROGRESS_UPDATE_SIGNAL,
    TRAFFIC_MONITOR_UPDATE_SIGNAL,
    COLLECTION_ADDED_SIGNAL,
    COLLECTION_REMOVED_SIGNAL,
    FILE_ADDED_SIGNAL,
    FILE_REMOVED_SIGNAL,
    FILE_DOWNLOAD_STATE_UPDATE_SIGNAL,
    FILE_STATE_CHANGED_SIGNAL,
    CONNECTION_CONNECTING_SIGNAL,
    CONNECTION_CONNECTED_SIGNAL,
    CONNECTION_DISCONNECT_SIGNAL,
    SCHEDULAR_STATE_CHANGED_SIGNAL,
    FATAL_ERROR_SIGNAL,
    WARNING_MESSAGE_SIGNAL,
    DEBUG_MESSAGE_SIGNAL,
    CONNECTION_LOST_SIGNAL,
    TASK_MOVED_SIGNAL,
    COLLECTION_MOVED_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(NNTPGrabCore, core, G_TYPE_OBJECT)

static void
core_init (NNTPGrabCore *obj)
{
    obj->socket_listener.socket_id = -1;
    obj->socket_command.socket_id = -1;
    memset(&obj->socket_listener.recv_buf, 0, sizeof(obj->socket_listener.recv_buf));
    memset(&obj->socket_command.recv_buf, 0, sizeof(obj->socket_command.recv_buf));
}

static void
core_finalize (GObject *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    if (core->socket_listener.socket_id > 0) {
        CLOSE(core->socket_listener.socket_id);
        core->socket_listener.socket_id = -1;
    }

    if (core->socket_command.socket_id > 0) {
        CLOSE(core->socket_command.socket_id);
        core->socket_command.socket_id = -1;
    }
}

static void
core_class_init (NNTPGrabCoreClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = core_finalize;

    signals[CONFIG_CHANGED_SIGNAL] =       g_signal_new("config_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__VOID,
                                                        G_TYPE_NONE, 0);

    signals[PART_DOWNLOAD_START_SIGNAL] = g_signal_new ("part_download_start",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_DONE_SIGNAL] =           g_signal_new ("part_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[PART_FAILED_SIGNAL] =         g_signal_new ("part_failed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT_BOOLEAN,
                                                        G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_BOOLEAN);

    signals[PART_PROGRESS_UPDATE_SIGNAL] = g_signal_new ("part_progress_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__STRING_INT_STRING_INT_INT_INT,
                                                         G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    signals[TRAFFIC_MONITOR_UPDATE_SIGNAL] = g_signal_new ("traffic_monitor_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__INT_INT_INT_INT_INT_INT_INT_INT_INT_INT_INT64_DOUBLE,
                                                         G_TYPE_NONE, 12, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT64, G_TYPE_DOUBLE);

    signals[COLLECTION_ADDED_SIGNAL] =    g_signal_new ("collection_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[COLLECTION_REMOVED_SIGNAL] =  g_signal_new ("collection_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[FILE_ADDED_SIGNAL] =          g_signal_new ("file_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_UINT64_UINT64_UINT64_UINT64_INT_INT_POINTER,
                                                        G_TYPE_NONE, 10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);

    signals[FILE_REMOVED_SIGNAL] =        g_signal_new ("file_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_UINT64_UINT64,
                                                        G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL] =  g_signal_new ("file_download_state_update",
                                                                G_OBJECT_CLASS_TYPE (klass),
                                                                G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                                0,
                                                                NULL, NULL,
                                                                nntpgrab_marshal_VOID__STRING_STRING_INT_INT_INT_UINT64_UINT64_UINT64_UINT64,
                                                                G_TYPE_NONE, 9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_STATE_CHANGED_SIGNAL] =  g_signal_new ("file_state_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[CONNECTION_CONNECTING_SIGNAL] =  g_signal_new   ("connection_connecting",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT,
                                                             G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

    signals[CONNECTION_CONNECTED_SIGNAL] =  g_signal_new    ("connection_connected",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                             G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    signals[CONNECTION_DISCONNECT_SIGNAL] = g_signal_new ("connection_disconnect",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_ENUM_STRING,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);


    signals[SCHEDULAR_STATE_CHANGED_SIGNAL] = g_signal_new ("schedular_state_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__INT_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);


    signals[FATAL_ERROR_SIGNAL] =           g_signal_new ("fatal_error",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[WARNING_MESSAGE_SIGNAL] =       g_signal_new ("warning_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[DEBUG_MESSAGE_SIGNAL] =         g_signal_new ("debug_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__POINTER_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_STRING);

    signals[CONNECTION_LOST_SIGNAL] =       g_signal_new ("connection_lost",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[TASK_MOVED_SIGNAL] =            g_signal_new ("task_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                          G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[COLLECTION_MOVED_SIGNAL] =      g_signal_new ("collection_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_INT,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);
}

static void
strip_newline(char *line)
{
    if (strlen(line) <= 1) {
        return;
    }

    if (line[strlen(line) - 1]  == '\n') {
        line[strlen(line) - 1] = '\0';
    }

    if (line[strlen(line) - 1]  == '\r') {
        line[strlen(line) - 1] = '\0';
    }

}

static gboolean
read_line(Connection *sock, char **ret_line, int timeout, gboolean *error_flag)
{
    int i;
    int len;
    char *line;
    time_t stamp;

    g_assert(sock);
    g_assert(ret_line);

    if (error_flag) {
        *error_flag = FALSE;
    }

    memset(sock->recv_buf + sock->recv_buf_len, 0, sizeof(sock->recv_buf) - sock->recv_buf_len);

    stamp = time(NULL);

    do {
        struct timeval tv;
        fd_set recv_fds;

        // Check is there is a newline in the buffer
        for (i = 0; i < sock->recv_buf_len; i++) {
            if (sock->recv_buf[i] == '\n') {
                // Newline found, split the buffer into a line and process it
                line = g_strndup(sock->recv_buf, i);

                strip_newline(line);

                memmove(sock->recv_buf, sock->recv_buf + i + 1, sock->recv_buf_len - i - 1);
                sock->recv_buf_len -= i + 1;

                *ret_line = line;

                return TRUE;
            }
        }

        // No newline found, try to read new data from the socket
        do {
            FD_ZERO(&recv_fds);
            FD_SET(sock->socket_id, &recv_fds);

            tv.tv_sec = timeout;
            tv.tv_usec = 0;

            len = select(sock->socket_id + 1, &recv_fds, NULL, NULL, &tv);
            if (len == 0) {
                // Read timeout
                *ret_line = g_strdup("Read timeout");
                return FALSE;
            } else if (len == 1) {
                gboolean beenhere = FALSE;

                do {
                    if (beenhere) {
                        g_usleep(G_USEC_PER_SEC / 10);
                    } else {
                        beenhere = TRUE;
                    }

                    len = recv(sock->socket_id, sock->recv_buf + sock->recv_buf_len, sizeof(sock->recv_buf) - sock->recv_buf_len, 0);
                } while (len == -1 && errno == EAGAIN && stamp + timeout > time(NULL));
            }
        } while (len == -1 && errno == EINTR && stamp + timeout > time(NULL));

        if (len == -1) {
            if (errno == EAGAIN) {
                // Read timeout
                *ret_line = g_strdup("Read timeout");
                return FALSE;
            }

            // Error occured
g_print(__FILE__":%i Error occured: %s\n", __LINE__, strerror(errno));
            if (error_flag) {
                *error_flag = TRUE;
            }

            CLOSE(sock->socket_id);
            sock->socket_id = -1;

            *ret_line = g_strdup_printf("Unexpected error occured while performing read: %s", strerror(errno));

            return FALSE;
        }

        if (len == 0) {
            // EOF, disconnect
g_print(__FILE__":%i EOF detected\n", __LINE__);
            if (error_flag) {
                *error_flag = TRUE;
            }

            CLOSE(sock->socket_id);
            sock->socket_id = -1;

            *ret_line = g_strdup(_("NNTPGrab server has disconnected"));

            return FALSE;
        }

        sock->recv_buf_len += len;
    } while (TRUE);

    g_assert_not_reached();

    return FALSE;
}

static gboolean
write_line(Connection *sock, const char *line)
{
    if (send(sock->socket_id, line, strlen(line), 0) == -1) {
//        imported_funcs.notify_disconnect(DISCONNECT_WRITE_ERROR, strerror(errno), data);

        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        return FALSE;
    }

    return TRUE;
}

static void
connect_to_server(Connection *sock, const char *hostname, int port, char **welcome_msg)
{
    struct addrinfo hints, *res, *ressave;
    int n;
    char *str_port;
    int last_errno;
    struct timeval tv;
#if 0
    SSL *ssl;
#endif

    g_assert(sock);
    g_assert(hostname);
    g_assert(port > 0);
    g_assert(port < 65535);

    if (sock->socket_id >= 0) {
        CLOSE(sock->socket_id);
    }

    sock->socket_id = -1;

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    str_port = g_strdup_printf("%i", port);
    n = getaddrinfo(hostname, str_port, &hints, &res);
    g_free(str_port);

    if (n != 0) {
        if (welcome_msg) {
            *welcome_msg = g_strdup(gai_strerror(n));
        }

        return;
    }

    ressave = res;
    last_errno = 0;

    while (res) {
        sock->socket_id = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((sock->socket_id == INVALID_SOCKET)) {
#else
        if ((sock->socket_id == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
            last_errno = errno;
            res = res->ai_next;
            sock->socket_id = -1;

            if (welcome_msg) {
                if (*welcome_msg) {
                    g_free(*welcome_msg);
                }

                *welcome_msg = g_strdup(strerror(last_errno));
            }

            continue;
        }

        // Set the connection timeout on the socket
        tv.tv_sec = 5;
        tv.tv_usec = 0;
        setsockopt(sock->socket_id, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(sock->socket_id, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));

        if (connect(sock->socket_id, res->ai_addr, (int) res->ai_addrlen) == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
        }

        // Connection could not be made, save the errno
#ifdef WIN32
        last_errno = WSAGetLastError();
#else
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else {
            last_errno = errno;
        }
#endif

        if (welcome_msg) {
            if (*welcome_msg) {
                g_free(*welcome_msg);
            }

#ifdef WIN32
            switch(WSAGetLastError()) {
                case WSAECONNREFUSED:
                    *welcome_msg = g_strdup(_("Connection refused"));
                    break;

                case WSAETIMEDOUT:
                    *welcome_msg = g_strdup(_("Connection timed out"));
                    break;

                default:
                    *welcome_msg = g_strdup_printf(_("Unknown error occured, WSAGetLastError() = %i"), WSAGetLastError());
                    break;
            }
#else
            *welcome_msg = g_strdup(strerror(last_errno));
#endif
        }

        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        res = res->ai_next;
    }

    if (sock->socket_id == -1) {
        freeaddrinfo(ressave);

        return;
    }

    freeaddrinfo(ressave);

#if 0
    if (use_ssl) {
        if ((ssl = prepare_ssl_connection(*conn_id, errmsg)) == NULL) {
            CLOSE(*conn_id);
            return NNTP_ERROR_SSL_INITIALISE;
        }
    } else {
        ssl = NULL;
    }
#endif

    if (!read_line(sock, welcome_msg, READ_TIMEOUT_VALUE, NULL)) {
        CLOSE(sock->socket_id);
        sock->socket_id = -1;

        return;
    }
}

/*********************************
 Callback functions
 *********************************/
static gboolean abort_flag = FALSE;
static GThread *thread_listener = NULL;

void emit_config_changed(NNTPGrabCore *core);

NNTPGrabCore *
get_core(void)
{
    return core;
}

static void
emit_schedular_state_changed(const char *command)
{
    SchedularState state;
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_STATE_CHANGED ([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);
        return;
    }

    if (!strncmp(values[1], "RUNNING", 7)) {
        state = SCHEDULAR_STATE_RUNNING;
    } else if (!strncmp(values[1], "STOPPING", 8)) {
        state = SCHEDULAR_STATE_STOPPING;
    } else if (!strncmp(values[1], "STOPPED", 7)) {
        state = SCHEDULAR_STATE_STOPPED;
    } else {
        char *msg;

        state = SCHEDULAR_STATE_STOPPED;
        msg = g_strdup_printf(__FILE__ ":%i Unknown schedular state: %s\n", __LINE__, values[1]);
        DEBUG(msg);
        g_free(msg);
    }

    nntpgrab_core_emit_schedular_state_changed(state, (values[2] && strlen(values[2]) > 0) ? values[2] : NULL);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_file_download_state_update(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("FILE_DOWNLOAD_STATE_UPDATE ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_file_download_state_update(values[1], values[2], atoi(values[3]), atoi(values[4]), atoi(values[5]), atoll(values[6]), atoll(values[7]), atoll(values[8]), atoll(values[9]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_file_state_changed(const char *command)
{
    TaskState old_state, new_state;
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("FILE_STATE_CHANGED ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    if (!strncmp(values[3], "WAITING_FOR_DOWNLOAD", 20)) {
        old_state = TASK_STATE_WAITING_FOR_DOWNLOAD;
    } else if (!strncmp(values[3], "DOWNLOADING", 11)) {
        old_state = TASK_STATE_DOWNLOADING;
    } else if (!strncmp(values[3], "WAITING_FOR_DECODE", 18)) {
        old_state = TASK_STATE_WAITING_FOR_DECODE;
    } else if (!strncmp(values[3], "DECODING", 8)) {
        old_state = TASK_STATE_DECODING;
    } else if (!strncmp(values[3], "FINISHED", 8)) {
        old_state = TASK_STATE_FINISHED;
    } else if (!strncmp(values[3], "SKIPPED", 7)) {
        old_state = TASK_STATE_SKIPPED;
    } else {
        char *msg;

        old_state = TASK_STATE_FINISHED;

        msg = g_strdup_printf(__FILE__ ":%i Unknown task state: %s\n", __LINE__, values[3]);
        DEBUG(msg);
        g_free(msg);
    }

    if (!strncmp(values[4], "WAITING_FOR_DOWNLOAD", 20)) {
        new_state = TASK_STATE_WAITING_FOR_DOWNLOAD;
    } else if (!strncmp(values[4], "DOWNLOADING", 11)) {
        new_state = TASK_STATE_DOWNLOADING;
    } else if (!strncmp(values[4], "WAITING_FOR_DECODE", 18)) {
        new_state = TASK_STATE_WAITING_FOR_DECODE;
    } else if (!strncmp(values[4], "DECODING", 8)) {
        new_state = TASK_STATE_DECODING;
    } else if (!strncmp(values[4], "FINISHED", 8)) {
        new_state = TASK_STATE_FINISHED;
    } else if (!strncmp(values[4], "SKIPPED", 7)) {
        new_state = TASK_STATE_SKIPPED;
    } else {
        char *msg;

        new_state = TASK_STATE_FINISHED;

        msg = g_strdup_printf(__FILE__ ":%i Unknown task state: %s\n", __LINE__, values[3]);
        DEBUG(msg);
        g_free(msg);
    }

    nntpgrab_core_emit_file_state_changed(values[1], values[2], old_state, new_state);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_traffic_monitor_update(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    int bytes_received[10];

    if (!re) {
        re = nntpgrab_utils_regex_compile("TRAFFIC_MONITOR_UPDATE ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9] && values[10] && values[11] && values[12])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    bytes_received[0] = atoi(values[1]);
    bytes_received[1] = atoi(values[2]);
    bytes_received[2] = atoi(values[3]);
    bytes_received[3] = atoi(values[4]);
    bytes_received[4] = atoi(values[5]);
    bytes_received[5] = atoi(values[6]);
    bytes_received[6] = atoi(values[7]);
    bytes_received[7] = atoi(values[8]);
    bytes_received[8] = atoi(values[9]);
    bytes_received[9] = atoi(values[10]);

    nntpgrab_core_emit_traffic_monitor_update(bytes_received, atoi(values[11]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_connection_connecting(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("CONNECTION_CONNECTING ([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_connecting(values[1], atoi(values[2]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_connection_connected(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("CONNECTION_CONNECTED ([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_connected(values[1], atoi(values[2]), values[3]);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_connection_disconnect(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    NNTPDisconnectType disconnect_type;

    if (!re) {
        re = nntpgrab_utils_regex_compile("CONNECTION_DISCONNECT ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    if (!strcmp(values[3], "NORMAL")) {
        disconnect_type = DISCONNECT_NORMAL;
    } else if (!strcmp(values[3], "NO_SUCH_HOST")) {
        disconnect_type = DISCONNECT_NO_SUCH_HOST;
    } else if (!strcmp(values[3], "CONNECTION_REFUSED")) {
        disconnect_type = DISCONNECT_CONNECTION_REFUSED;
    } else if (!strcmp(values[3], "TOO_MANY_CONNECTIONS")) {
        disconnect_type = DISCONNECT_TOO_MANY_CONNECTIONS;
    } else if (!strcmp(values[3], "CONNECT_TIMEOUT")) {
        disconnect_type = DISCONNECT_CONNECT_TIMEOUT;
    } else if (!strcmp(values[3], "READ_ERROR")) {
        disconnect_type = DISCONNECT_READ_ERROR;
    } else if (!strcmp(values[3], "READ_TIMEOUT")) {
        disconnect_type = DISCONNECT_READ_TIMEOUT;
    } else if (!strcmp(values[3], "WRITE_ERROR")) {
        disconnect_type = DISCONNECT_WRITE_ERROR;
    } else if (!strcmp(values[3], "IDLE_TIMEOUT")) {
        disconnect_type = DISCONNECT_IDLE_TIMEOUT;
    } else if (!strcmp(values[3], "INVALID_MSG")) {
        disconnect_type = DISCONNECT_INVALID_MSG;
    } else if (!strcmp(values[3], "LOGIN_FAILURE")) {
        disconnect_type = DISCONNECT_LOGIN_FAILURE;
    } else if (!strcmp(values[3], "UNEXPECTED")) {
        disconnect_type = DISCONNECT_UNEXPECTED;
    } else {
        char *msg;

        disconnect_type = DISCONNECT_UNEXPECTED;

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);
    }

    nntpgrab_core_emit_disconnect(values[1], atoi(values[2]), disconnect_type, values[4]);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_part_download_start(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("PART_DOWNLOAD_START ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_part_download_start(values[1], atoi(values[2]), values[3], values[4], atoi(values[5]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_part_done(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("PART_DONE ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_part_done(values[1], atoi(values[2]), values[3], values[4], atoi(values[5]), atoi(values[6]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_part_failed(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    gboolean all_servers_tried;

    if (!re) {
        re = nntpgrab_utils_regex_compile("PART_FAILED ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    if (!strcmp(values[7], "TRUE")) {
        all_servers_tried = TRUE;
    } else {
        all_servers_tried = FALSE;
    }

    nntpgrab_core_emit_part_failed(values[1], atoi(values[2]), values[3], values[4], atoi(values[5]), atoi(values[6]), all_servers_tried);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_part_progress_update(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("PART_PROGRESS_UPDATE ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_part_progress_update(values[1], atoi(values[2]), values[3], atoi(values[4]), atoi(values[5]), atoi(values[6]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_file_added(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    char **values_groups;
    int i;
    NNTPFile file;
    NNTPCollection collection;

    if (!re) {
        re = nntpgrab_utils_regex_compile("FILE_ADDED ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9] && values[10])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    memset(&collection, 0, sizeof(collection));
    memset(&file, 0, sizeof(file));
    strncpy((char*) collection.collection_name, values[1], sizeof(collection.collection_name));
    strncpy((char*) file.subject, values[2], sizeof(file.subject));
    strncpy((char*) file.poster, values[3], sizeof(file.poster));
    file.stamp = atoi(values[4]);
    file.file_size = atoll(values[5]);
    collection.total_size = atoll(values[6]);
    collection.total_size_remaining = atoll(values[7]);
    file.numparts = atoi(values[9]);

    if (!strcmp(values[8], "WAITING_FOR_DOWNLOAD")) {
        file.status = TASK_STATE_WAITING_FOR_DOWNLOAD;
    } else if (!strcmp(values[8], "DOWNLOADING")) {
        file.status = TASK_STATE_DOWNLOADING;
    } else if (!strcmp(values[8], "WAITING_FOR_DECODE")) {
        file.status = TASK_STATE_WAITING_FOR_DECODE;
    } else if (!strcmp(values[8], "DECODING")) {
        file.status = TASK_STATE_DECODING;
    } else if (!strcmp(values[8], "FINISHED")) {
        file.status = TASK_STATE_FINISHED;
    } else if (!strcmp(values[8], "SKIPPED")) {
        file.status = TASK_STATE_SKIPPED;
    } else {
        char *msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        nntpgrab_utils_regex_matches_free(values);
        return;
    }

    values_groups = g_strsplit(values[10], "|", 0);
    i = 0;
    while (values_groups[i]) {
        file.groups = g_list_append(file.groups, values_groups[i]);
        i++;
    }

    nntpgrab_core_emit_file_added(&collection, &file);

    g_list_free(file.groups);
    g_strfreev(values_groups);
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_file_removed(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("FILE_REMOVED ([^\t]*)\t([^\t]*)\t([^\t]*)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_file_removed(values[1], values[2], atoll(values[3]), atoll(values[4]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_task_moved(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("TASK_MOVED ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_task_moved(values[1], values[2], values[3], atoi(values[4]), atoi(values[5]));
    nntpgrab_utils_regex_matches_free(values);
}

static void
emit_collection_moved(const char *command)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("COLLECTION_MOVED ([^\t]+)\t([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3])) {
        char *msg;

        nntpgrab_utils_regex_matches_free(values);

        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, command);
        DEBUG(msg);
        g_free(msg);

        return;
    }

    nntpgrab_core_emit_collection_moved(values[1], atoi(values[2]), atoi(values[3]));
    nntpgrab_utils_regex_matches_free(values);
}

static gboolean
do_emit_connection_lost(gpointer data)
{
    char *errmsg = (char *) data;

    if (!get_core()) {
        g_free(errmsg);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "connection_lost", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
emit_connection_lost(const char *errmsg)
{
    if (!get_core()) {
        return;
    }

    g_idle_add(do_emit_connection_lost, (gpointer) g_strdup(errmsg));
}

static gboolean
has_signal_handler_pending(int signal_id)
{
    return g_signal_has_handler_pending(core, signal_id, 0, TRUE);
}

static void
process_message(const char *line)
{
    char *ptr;
    char *command;
    unsigned int hash;

    ptr = strstr(line, " ");
    if (ptr) {
        command = g_strndup(line, ptr - line);
    } else {
        command = g_strdup(line);
    }

    // Generate a hash of the message to optimize the processing
    hash = g_str_hash(command);

    switch(hash) {
        case 1903602327U:
            if (!strcmp(command, "CONFIG_CHANGED")) {
                if (has_signal_handler_pending(signals[CONFIG_CHANGED_SIGNAL])) {
                    nntpgrab_core_emit_config_changed();
                }
                break;
            }

        case 1349955479U:
            if (!strncmp(command, "PART_DOWNLOAD_START", 19)) {
                if (has_signal_handler_pending(signals[PART_DOWNLOAD_START_SIGNAL])) {
                    emit_part_download_start(line);
                }
                break;
            }

        case 1006630990U:
            if (!strncmp(command, "PART_DONE", 9)) {
                if (has_signal_handler_pending(signals[PART_DONE_SIGNAL])) {
                    emit_part_done(line);
                }
                break;
            }

        case 1048928777U:
            if (!strncmp(command, "PART_FAILED", 11)) {
                if (has_signal_handler_pending(signals[PART_FAILED_SIGNAL])) {
                    emit_part_failed(line);
                }
                break;
            }

        case 424842768U:
            if (!strncmp(command, "TRAFFIC_MONITOR_UPDATE", 22)) {
                if (has_signal_handler_pending(signals[TRAFFIC_MONITOR_UPDATE_SIGNAL])) {
                    emit_traffic_monitor_update(line);
                }
                break;
            }

        case 1761576655U:
            if (!strncmp(command, "PART_PROGRESS_UPDATE", 20)) {
                if (has_signal_handler_pending(signals[PART_PROGRESS_UPDATE_SIGNAL])) {
                    emit_part_progress_update(line);
                }
                break;
            }

        case 3983341567U:
            if (!strncmp(command, "COLLECTION_ADDED", 16)) {
                if (has_signal_handler_pending(signals[COLLECTION_ADDED_SIGNAL])) {
                    g_assert(strlen(line) > 17);
                    nntpgrab_core_emit_collection_added(line + 17);
                }
                break;
            }

        case 3415304095U:
            if (!strncmp(command, "COLLECTION_REMOVED", 18)) {
                if (has_signal_handler_pending(signals[COLLECTION_REMOVED_SIGNAL])) {
                    g_assert(strlen(line) > 19);
                    nntpgrab_core_emit_collection_removed(line + 19);
                }
                break;
            }

        case 3528469597U:
            if (!strncmp(command, "FILE_ADDED", 10)) {
                if (has_signal_handler_pending(signals[FILE_ADDED_SIGNAL])) {
                    emit_file_added(line);
                }
                break;
            }

        case 75037821U:
            if (!strncmp(command, "FILE_REMOVED", 12)) {
                if (has_signal_handler_pending(signals[FILE_REMOVED_SIGNAL])) {
                    emit_file_removed(line);
                }
                break;
            }

        case 2847286315U:
            if (!strncmp(command, "FILE_DOWNLOAD_STATE_UPDATE", 26)) {
                if (has_signal_handler_pending(signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL])) {
                    emit_file_download_state_update(line);
                }
                break;
            }

        case 1809569603U:
            if (!strncmp(command, "FILE_STATE_CHANGED", 18)) {
                if (has_signal_handler_pending(signals[FILE_STATE_CHANGED_SIGNAL])) {
                    emit_file_state_changed(line);
                }
                break;
            }

        case 581271225U:
            if (!strncmp(command, "CONNECTION_CONNECTING", 21)) {
                if (has_signal_handler_pending(signals[CONNECTION_CONNECTING_SIGNAL])) {
                    emit_connection_connecting(line);
                }
                break;
            }

        case 3898075848U:
            if (!strncmp(command, "CONNECTION_CONNECTED", 20)) {
                if (has_signal_handler_pending(signals[CONNECTION_CONNECTED_SIGNAL])) {
                    emit_connection_connected(line);
                }
                break;
            }

        case 1887328413U:
            if (!strncmp(command, "CONNECTION_DISCONNECT", 21)) {
                if (has_signal_handler_pending(signals[CONNECTION_DISCONNECT_SIGNAL])) {
                    emit_connection_disconnect(line);
                }
                break;
            }


        case 2671763622U:
            if (!strncmp(command, "SCHEDULAR_STATE_CHANGED", 23)) {
                if (has_signal_handler_pending(signals[SCHEDULAR_STATE_CHANGED_SIGNAL])) {
                    emit_schedular_state_changed(line);
                }
                break;
            }

        case 507002573U:
            if (!strncmp(command, "FATAL_ERROR", 11)) {
                if (has_signal_handler_pending(signals[CONNECTION_LOST_SIGNAL])) {
                    g_assert(strlen(line) > 12);
                    emit_connection_lost(line + 13);
                }
                break;
            }

        case 2480026148U:
            if (!strncmp(command, "WARNING_MESSAGE", 15)) {
                if (has_signal_handler_pending(signals[WARNING_MESSAGE_SIGNAL])) {
                    g_assert(strlen(line) > 16);
                    nntpgrab_core_emit_warning(line + 16);
                }
                break;
            }

        case 4232637915U:
            if (!strncmp(command, "DEBUG_MESSAGE", 13)) {
                if (has_signal_handler_pending(signals[DEBUG_MESSAGE_SIGNAL])) {
                    g_assert(strlen(line) > 14);
                    nntpgrab_core_emit_debug_message(line + 14);
                }
                break;
            }

        case 2708266841U:
            if (!strncmp(command, "TASK_MOVED", 10)) {
                if (has_signal_handler_pending(signals[TASK_MOVED_SIGNAL])) {
                    emit_task_moved(line);
                }
                break;
            }

        case 3994768818U:
            if (!strncmp(command, "COLLECTION_MOVED", 16)) {
                if (has_signal_handler_pending(signals[COLLECTION_MOVED_SIGNAL])) {
                    emit_collection_moved(line);
                }
                break;
            }

        default:
            {
                char *msg;

                msg = g_strdup_printf("Unknown data received: %s\n", line);
                DEBUG(msg);
                g_free(msg);

                msg = g_strdup_printf("hash = %u\n", hash);
                DEBUG(msg);
                g_free(msg);
            }

            break;
    };

    g_free(command);

    // Give other threads some time to handle the signals
    g_thread_yield();
}

static gpointer
callbacks_thread(gpointer data)
{
    NNTPGrabCore *core = data;

    do {
        gboolean error_flag;
        char *line = NULL;

        if (!read_line(&core->socket_listener, &line, 1, &error_flag)) {
            if (error_flag) {
                // An error occured!
                char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s", __LINE__, line);

                emit_connection_lost(msg);
                DEBUG(msg);

                g_free(line);
                g_free(msg);

                return NULL;
            }

            if (line) {
                g_free(line);
            }

            // No data received, sleep
            g_usleep(G_USEC_PER_SEC / 10);

            continue;
        }

        process_message(line);

        g_free(line);
    } while (!abort_flag);

    return NULL;
}

static void
initialize_callbacks(NNTPGrabCore *core)
{
    char *line = NULL;

    abort_flag = FALSE;
    write_line(&core->socket_listener, "MODE LISTENER\n");
    if (!read_line(&core->socket_listener, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        return;
    }

    if (strcmp(line, "OK")) {
        char *msg = g_strdup_printf(__FILE__ ":%i unknown response received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        return;
    }

    g_free(line);

    thread_listener = g_thread_create(callbacks_thread, core, TRUE, NULL);

    g_assert(thread_listener != NULL);
}

static void
destroy_callbacks(NNTPGrabCore *core)
{
    abort_flag = TRUE;
    if (thread_listener) {
        g_thread_join(thread_listener);
        thread_listener = NULL;
    }
}

/*********************************
 Public functions
 *********************************/
static GStaticMutex command_mutex = G_STATIC_MUTEX_INIT;

NNTPGrabCore *
nntpgrab_networked_init(void)
{
#ifdef WIN32
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
#endif

    g_type_init();
    if (!g_thread_supported ()) g_thread_init (NULL);

#ifdef WIN32
    // Initialise Winsock
    wVersionRequested = MAKEWORD( 2, 2 );

    err = WSAStartup( wVersionRequested, &wsaData );
    if ( err != 0 ) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        g_error(_("Winsock could not be initialised"));
        return FALSE;
    }

    /* Confirm that the WinSock DLL supports 2.2.*/
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */

    if ( LOBYTE( wsaData.wVersion ) != 2 ||
         HIBYTE( wsaData.wVersion ) != 2 ) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup( );

        g_error(_("Winsock 2.2 or higher is required for this program"));
        return FALSE;
    }
#endif

    core = g_object_new(NNTPGRAB_TYPE_CORE, NULL);

    return core;
}

gboolean
nntpgrab_networked_connect(NNTPGrabCore *obj, const char *hostname, int port, const char *username, const char *password, gboolean use_ssl, char **err)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    int api_version = 0;

    g_assert(core != NULL);

    if (err) {
        *err = NULL;
    }

    connect_to_server(&core->socket_listener, hostname, port, &line);
    if (core->socket_listener.socket_id < 0) {
        if (err) {
            if (line) {
                *err = g_strdup(line);
            } else {
                *err = g_strdup_printf(__FILE__ ":%i connect_to_server() FAILED", __LINE__);
            }
        }

        if (line) {
            g_free(line);
        }

        return FALSE;
    }

    g_assert(line);

    // API check
    if (sscanf(line, "NNTPGrab Server - API version %i", &api_version) != 1) {
        if (err) {
            *err = g_strdup_printf(__FILE__ ":%i Unknown response received from server:\n%s\n", __LINE__, line);
        }

        g_free(line);

        return FALSE;
    }

    g_free(line);

    if (api_version != NNTPGRAB_API_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Server API mismatch (Server API version = %i, frontend's version = %i)"), NNTPGRAB_API_VERSION, api_version);
        }

        return FALSE;
    }

    initialize_callbacks(core);

    g_static_mutex_lock(&command_mutex);

    connect_to_server(&core->socket_command, hostname, port, &line);
    if (core->socket_command.socket_id < 0) {
        if (err) {
            *err = g_strdup_printf(__FILE__ ":%i %s", __LINE__, (line ? line : ""));
        }

        if (line) {
            g_free(line);
        }

        g_static_mutex_unlock(&command_mutex);

        return FALSE;
    }

    g_free(line);

    write_line(&core->socket_command, "MODE COMMAND\n");
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (strcmp(line, "OK")) {
        char *msg = g_strdup_printf(__FILE__ ":%i unknown response received: %s\n", __LINE__, (line ? line : ""));
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    g_free(line);

    g_static_mutex_unlock(&command_mutex);

    return TRUE;
}

gboolean
nntpgrab_networked_get_is_connected(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    if (core->socket_listener.socket_id < 0) {
        return FALSE;
    }

    if (core->socket_command.socket_id < 0) {
        return FALSE;
    }

    return TRUE;
}

void
nntpgrab_networked_cleanup(NNTPGrabCore *obj)
{
    destroy_callbacks(obj);
    g_object_unref(obj);
}

GList *
nntpgrab_config_get_avail_servers(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    GList *ret = NULL;

    g_static_mutex_lock(&command_mutex);

    write_line(&core->socket_command, "CONFIG_GET_AVAIL_SERVERS\n");

    while (TRUE) {
        if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            GList *list;
            char *msg;

            msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_free(line);

            list = ret;
            while (list) {
                g_free(list->data);
                list = g_list_next(list);
            }
            g_list_free(ret);

            g_static_mutex_unlock(&command_mutex);
            return NULL;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "NULL", 4)) {
            break;
        }

        ret = g_list_append(ret, line);
    }

    g_static_mutex_unlock(&command_mutex);

    return ret;
}

void
nntpgrab_config_free_avail_servers(NNTPGrabCore *obj, GList *servers)
{
    GList *list;

    list = servers;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    g_list_free(servers);
}

gboolean
nntpgrab_config_get_server_info(NNTPGrabCore *obj, const char *servername, ConfigServer *ret)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    ConfigServer server;

    g_assert(servername);
    g_assert(ret);

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_GET_SERVER_INFO %s\n", servername);
    write_line(&core->socket_command, line);
    g_free(line);

    // Servername
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "NULL", 4)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.servername, line, sizeof(server.servername));
    g_free(line);

    // Hostname
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.hostname, line, sizeof(server.hostname));
    g_free(line);

    // Port
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.port = atoi(line);
    g_free(line);

    // Username
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.username, line, sizeof(server.username));
    g_free(line);

    // Password
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    strncpy(server.password, line, sizeof(server.password));
    g_free(line);

    // Max threads
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.max_threads = atoi(line);
    g_free(line);

    // Priority
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    server.priority = atoi(line);
    g_free(line);

    // Use SSL
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        server.use_ssl = TRUE;
    } else {
        server.use_ssl = FALSE;
    }

    g_free(line);

    // Enabled
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        server.enabled = TRUE;
    } else {
        server.enabled = FALSE;
    }

    g_free(line);

    // NULL
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "NULL", 4)) {
        g_free(line);
        g_static_mutex_unlock(&command_mutex);

        memcpy(ret, &server, sizeof(ConfigServer));
        return TRUE;
    }

    g_free(line);

    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_add_server(NNTPGrabCore *obj, ConfigServer new_server)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *use_ssl_str;
    char *enabled_str;
    char *msg;

    g_assert(strlen(new_server.servername) > 0);
    g_assert(strlen(new_server.hostname) > 0);

    g_static_mutex_lock(&command_mutex);

    if (new_server.use_ssl) {
        use_ssl_str = "TRUE";
    } else {
        use_ssl_str = "FALSE";
    }

    if (new_server.enabled) {
        enabled_str = "TRUE";
    } else {
        enabled_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_ADD_SERVER %s\t%s\t%i\t%s\t%s\t%i\t%i\t%s\t%s\n", new_server.servername, new_server.hostname, new_server.port, new_server.username, new_server.password, new_server.max_threads, new_server.priority, use_ssl_str, enabled_str);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_del_server(NNTPGrabCore *obj, const char *servername)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    g_assert(servername);

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_DEL_SERVER %s\n", servername);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_config_edit_server(NNTPGrabCore *obj, const char *servername, ConfigServer new_server)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *use_ssl_str;
    char *enabled_str;
    char *msg;

    g_assert(servername);
    g_assert(strlen(new_server.servername) > 0);
    g_assert(strlen(new_server.hostname) > 0);

    g_static_mutex_lock(&command_mutex);

    if (new_server.use_ssl) {
        use_ssl_str = "TRUE";
    } else {
        use_ssl_str = "FALSE";
    }

    if (new_server.enabled) {
        enabled_str = "TRUE";
    } else {
        enabled_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_EDIT_SERVER %s\t%s\t%s\t%i\t%s\t%s\t%i\t%i\t%s\t%s\n", servername, new_server.servername, new_server.hostname, new_server.port, new_server.username, new_server.password, new_server.max_threads, new_server.priority, use_ssl_str, enabled_str);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

ConfigOpts
nntpgrab_config_get_opts(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    ConfigOpts opts;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    memset(&opts, 0, sizeof(opts));

    write_line(&core->socket_command, "CONFIG_GET_OPTS\n");

    // Download directory
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    strncpy(opts.download_directory, line, sizeof(opts.download_directory));
    g_free(line);

    // Temp directory
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    strncpy(opts.temp_directory, line, sizeof(opts.temp_directory));
    g_free(line);

    // Enable intelligent PAR2 downloading
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "FALSE", 5)) {
        opts.enable_intelligent_par2_downloading = FALSE;
    } else if (!strncmp(line, "TRUE", 4)) {
        opts.enable_intelligent_par2_downloading = TRUE;
    } else {
        msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
    }

    g_free(line);

    // NULL
    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return opts;
    }

    if (!strncmp(line, "NULL", 4)) {
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return opts;
    }

    g_static_mutex_unlock(&command_mutex);

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);

    return opts;
}

void
nntpgrab_config_set_opts(NNTPGrabCore *obj, ConfigOpts opts)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *enable_par2_str;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    if (opts.enable_intelligent_par2_downloading) {
        enable_par2_str = "TRUE";
    } else {
        enable_par2_str = "FALSE";
    }

    line = g_strdup_printf("CONFIG_SET_OPTS %s\t%s\t%s\n", opts.download_directory, opts.temp_directory, enable_par2_str);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return;
}

gboolean
nntpgrab_config_get_folder_listing(NNTPGrabCore *obj, const char *parent, GList **folders)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char **values;
    ConfigFolder *folder;

    g_assert(folders);
    *folders = NULL;

    g_static_mutex_lock(&command_mutex);

    line = g_strdup_printf("CONFIG_GET_FOLDER_LISTING %s\n", (parent ? parent : ""));
    write_line(&core->socket_command, line);
    g_free(line);

    do {
        if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_free(line);
            g_static_mutex_unlock(&command_mutex);
            return FALSE;
        }

        if (!strncmp(line, "FALSE", 5)) {
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return FALSE;
        }

        if (!strncmp(line, "NULL", 4)) {
            g_free(line);
            break;
        }

        values = g_strsplit(line, "\t", 0);
        if (values && values[0] && values[1]) {
            folder = g_malloc0(sizeof(ConfigFolder));

            strncpy(folder->folder, values[0], sizeof(folder->folder) - 1);
            if (!strcmp(values[1], "TRUE")) {
                folder->has_subfolders = TRUE;
            } else {
                folder->has_subfolders = FALSE;
            }

            *folders = g_list_append(*folders, folder);
        }

        if (values) {
            g_strfreev(values);
        }

        g_free(line);
    } while (TRUE);

    g_static_mutex_unlock(&command_mutex);

    return TRUE;
}

void
nntpgrab_config_free_folder_listing(NNTPGrabCore *obj, GList *folders)
{
    GList *list = folders;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }
    g_list_free(folders);
}

gboolean
nntpgrab_schedular_start(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    write_line(&core->socket_command, "SCHEDULAR_START\n");

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_stop(NNTPGrabCore *obj, gboolean wait)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *wait_str;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    if (wait) {
        wait_str = "TRUE";
    } else {
        wait_str = "FALSE";
    }

    line = g_strdup_printf("SCHEDULAR_STOP %s\n", wait_str);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

SchedularState
nntpgrab_schedular_get_state(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    write_line(&core->socket_command, "SCHEDULAR_GET_STATE\n");

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "RUNNING", 7)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_RUNNING;
    }

    if (!strncmp(line, "STOPPING", 8)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_STOPPING;
    }

    if (!strncmp(line, "STOPPED", 7)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return SCHEDULAR_STATE_STOPPED;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return SCHEDULAR_STATE_STOPPED;
}

gboolean
nntpgrab_schedular_add_task_to_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, guint64 file_size, GList *groups, GList *parts, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    GList *list;
    char *groups_str = NULL;
    char *parts_str = NULL;
    char *msg;

    g_assert(collection_name);
    g_assert(subject);
    g_assert(poster);
    g_assert(groups);
    g_assert(parts);

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    list = groups;
    while (list) {
        if (!groups_str) {
            groups_str = g_strdup(list->data);
        } else {
            char *tmp = g_strdup_printf("%s|%s", groups_str, (char*) list->data);
            g_free(groups_str);
            groups_str = tmp;
        }

        list = g_list_next(list);
    }

    list = parts;
    while (list) {
        NNTPGrabPart *segment = list->data;

        if (!parts_str) {
            parts_str = g_strdup_printf("%s,%i,%i", segment->message_id, segment->partnum, segment->size);
        } else {
            char *tmp = g_strdup_printf("%s|%s,%i,%i", parts_str, segment->message_id, segment->partnum, segment->size);
            g_free(parts_str);
            parts_str = tmp;
        }

        list = g_list_next(list);
    }

#if SIZEOF_LONG == 8
    line = g_strdup_printf("SCHEDULAR_ADD_TASK_TO_QUEUE %s\t%s\t%s\t%li\t%li\t%s\t%s\n", collection_name, subject, poster, stamp, file_size, groups_str, parts_str);
#else
    line = g_strdup_printf("SCHEDULAR_ADD_TASK_TO_QUEUE %s\t%s\t%s\t%lli\t%lli\t%s\t%s\n", collection_name, subject, poster, (guint64) stamp, file_size, groups_str, parts_str);
#endif
    write_line(&core->socket_command, line);
    g_free(line);
    g_free(groups_str);
    g_free(parts_str);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("Connection lost to the NNTPGrab Server: %s\n"), line);
        }
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);

        if (errmsg) {
            *errmsg = g_strdup_printf(_("INVALID_COMMAND received\n"));
        }
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_del_task_from_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    // There is no assertion on the subject because the user can send a request
    // to remove all the items in the given collection

    line = g_strdup_printf("SCHEDULAR_DEL_TASK_FROM_QUEUE %s\t%s\n", collection_name, (subject ? subject : ""));
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_restart_task(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    // There is no assertion on the subject because the user can send a request
    // to remove all the items in the given collection

    line = g_strdup_printf("SCHEDULAR_RESTART_TASK %s\t%s\n", collection_name, (subject ? subject : ""));
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_save_queue(NNTPGrabCore *obj, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    if (errmsg) {
        *errmsg = NULL;
    }

    g_static_mutex_lock(&command_mutex);

    write_line(&core->socket_command, "SCHEDULAR_SAVE_QUEUE\n");

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        if (strlen(line) > 6 && errmsg) {
            *errmsg = g_strdup(line + 6);
        }

        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

void
nntpgrab_schedular_foreach_task(NNTPGrabCore *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, gpointer data)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *collection_name = NULL;

    g_static_mutex_lock(&command_mutex);

    write_line(&core->socket_command, "SCHEDULAR_FOREACH_TASK\n");

    do {
        if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
            char *msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
            DEBUG(msg);
            g_free(msg);
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return;
        }

        if (!strncmp(line, "INVALID_COMMAND", 15)) {
            char *msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
            DEBUG(msg);
            g_free(msg);
            g_static_mutex_unlock(&command_mutex);
            g_free(line);
            return;
        }

        if (!strncmp(line, "NULL", 4)) {
            g_free(line);
            break;
        }

        g_assert(strlen(line) > 1);

        // Collection
        if (!strncmp(line, "COLLECTION", 10)) {
            char **values;
            guint64 file_size;
            guint64 file_size_remaining;

            values = g_strsplit(line + 11, "\t", 0);

            if (!(values && values[0] && values[1] && values[2] && values[3])) {
                char *msg;

                if (values) {
                    g_strfreev(values);
                }

                if (collection_name) {
                    g_free(collection_name);
                }

                msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
                DEBUG(msg);
                g_free(msg);

                g_free(line);
                g_static_mutex_unlock(&command_mutex);
                return;
            }

            file_size = g_ascii_strtoull(values[1], (char**) NULL, 10);
            file_size_remaining = g_ascii_strtoull(values[2], (char**) NULL, 10);

            g_free(line);

            if (collection_func) {
                collection_func(values[0], file_size, file_size_remaining, atoi(values[3]), data);
            }

            collection_name = g_strdup(values[0]);

            g_strfreev(values);

            continue;
        }

        // File
        if (line[0] == '\t') {
            char **values;
            TaskState status;
            guint64 file_size;
            guint64 file_size_remaining;
            gint64 stamp;

            g_assert(collection_name != NULL);

            values = g_strsplit(line + 6, "\t", 0);

            if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9] && values[10] && values[11])) {
                char *msg;

                if (values) {
                    g_strfreev(values);
                }

                msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
                DEBUG(msg);
                g_free(msg);

                g_free(collection_name);
                g_free(line);
                g_static_mutex_unlock(&command_mutex);
                return;
            }

            if (!strcmp(values[9], "WAITING_FOR_DOWNLOAD")) {
                status = TASK_STATE_WAITING_FOR_DOWNLOAD;
            } else if (!strcmp(values[9], "DOWNLOADING")) {
                status = TASK_STATE_DOWNLOADING;
            } else if (!strcmp(values[9], "WAITING_FOR_DECODE")) {
                status = TASK_STATE_WAITING_FOR_DECODE;
            } else if (!strcmp(values[9], "DECODING")) {
                status = TASK_STATE_DECODING;
            } else if (!strcmp(values[9], "FINISHED")) {
                status = TASK_STATE_FINISHED;
            } else if (!strcmp(values[9], "SKIPPED")) {
                status = TASK_STATE_SKIPPED;
            } else {
                status = TASK_STATE_SKIPPED;
            }

#if GLIB_CHECK_VERSION(2,12,0)
            file_size = g_ascii_strtoull(values[3], (char **) NULL, 10);
            file_size_remaining = g_ascii_strtoull(values[4], (char **) NULL, 10);
            stamp = g_ascii_strtoll(values[2], (char **) NULL, 10);
#else
            file_size = strtoull(values[3], (char **) NULL, 10);
            file_size_remaining = strtoull(values[4], (char **) NULL, 10);
            stamp = strtoll(values[2], (char **) NULL, 10);
#endif

            if (file_func) {
                file_func(collection_name, values[0], values[1], stamp, file_size, file_size_remaining, atoi(values[5]), atoi(values[6]), atoi(values[7]), atoi(values[8]), status, values[10], data);
            }

            if (group_func) {
                char **values_groups = g_strsplit(values[11], "|", 0);
                int i = 0;
                while (values_groups[i]) {
                    group_func(collection_name, values[0], values_groups[i], data);
                    i++;
                }
                g_strfreev(values_groups);
            }

            g_strfreev(values);
        }

        g_free(line);
    } while(TRUE);

    if (collection_name) {
        g_free(collection_name);
    }

    g_static_mutex_unlock(&command_mutex);
}

gboolean
nntpgrab_schedular_move_task(NNTPGrabCore *obj, const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name_src);
    g_assert(collection_name_dest);
    g_assert(position_dest >= -1);

    // There is no assertion on the subject because the user can send a request
    // to move the given collection itself

    line = g_strdup_printf("SCHEDULAR_MOVE_TASK %s\t%s\t%s\t%i\n", collection_name_src, (subject_src ? subject_src : ""), collection_name_dest, position_dest);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}

gboolean
nntpgrab_schedular_move_collection(NNTPGrabCore *obj, const char *collection_name, int new_position)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *line = NULL;
    char *msg;

    g_static_mutex_lock(&command_mutex);

    g_assert(collection_name);
    g_assert(new_position >= -1);

    line = g_strdup_printf("SCHEDULAR_MOVE_COLLECTION %s\t%i\n", collection_name, new_position);
    write_line(&core->socket_command, line);
    g_free(line);

    if (!read_line(&core->socket_command, &line, READ_TIMEOUT_VALUE, NULL)) {
        msg = g_strdup_printf(__FILE__ ":%i Connection lost to NNTPGrab Server: %s\n", __LINE__, line);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "INVALID_COMMAND", 15)) {
        msg = g_strdup_printf(__FILE__ ":%i INVALID_COMMAND received\n", __LINE__);
        DEBUG(msg);
        g_free(msg);
        g_free(line);
        g_static_mutex_unlock(&command_mutex);
        return FALSE;
    }

    if (!strncmp(line, "FALSE", 5)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return FALSE;
    }

    if (!strncmp(line, "TRUE", 4)) {
        g_static_mutex_unlock(&command_mutex);
        g_free(line);
        return TRUE;
    }

    msg = g_strdup_printf(__FILE__ ":%i Unknown reply received: %s\n", __LINE__, line);
    DEBUG(msg);
    g_free(msg);

    g_free(line);
    g_static_mutex_unlock(&command_mutex);

    return FALSE;
}
