/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include "gui.h"
#include "config_gui.h"
#if GTK_CHECK_VERSION(2,12,0)
#include <gtkhtml/gtkhtml.h>
#else
#include <glade/glade.h>
#endif
#include "donate.xpm"
#include "icon_64x64.xpm"

#ifdef DARWIN
#include <ige-mac-menu.h>
#endif

#ifdef WIN32
#include <windows.h>
#endif

#if GTK_CHECK_VERSION(2,12,0)
static GtkBuilder *builder = NULL;
#else
static GladeXML *xml = NULL;
#endif
NNTPGrabCore *core = NULL;

void debug_initialize(void);
void info_initialize(void);
void config_initialize(void);
void config_gui_initialize(void);
void config_display_initialize(void);
void queue_initialize(void);
void callbacks_initialize(void);
void nzb_initialize(void);
void connections_initialize(void);
void notify_initialize(void);
void monitor_initialize(void);
void par2_initialize(void);
#if GTK_CHECK_VERSION(2,12,0)
void initialize_html_widget(void);
#endif

gboolean    ipc_initialize(void);
void        ipc_cleanup(void);
void        ipc_send_open_nzb_file_request(const char *nzb_file);
gboolean    ipc_test_is_nntpgrab_already_running(void);

#if GTK_CHECK_VERSION(2,10,0)
gboolean    tray_icon_get_is_shown(void);
#endif

static const char *nzb_file = NULL;

G_MODULE_EXPORT void
on_btnQuit_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowMain = get_widget("windowMain");

    gtk_widget_hide(windowMain);
    gtk_main_quit();
}

#if GTK_CHECK_VERSION(2,12,0)
GObject *
get_object(const char *name)
{
    GObject *obj;

    if (!builder) {
        return NULL;
    }

    obj = gtk_builder_get_object(builder, name);
    if (!obj) {
        g_error(_("No widget found with name '%s'\n"), name);
    }

    return obj;
}

GtkWidget *
get_widget(const char *name)
{
    return GTK_WIDGET(get_object(name));
}

#else

GtkWidget *
get_widget(const char *name)
{
    GtkWidget *widget;

    if (!xml) {
        return NULL;
    }

    widget = glade_xml_get_widget(xml, name);
    if (!widget) {
        g_error(_("No widget found with name '%s'\n"), name);
    }

    return widget;
}
#endif

static void
log_func (const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer user_data)
{
    char *msg;
    GtkWidget *window;

    window = get_widget("windowMain");
    if (log_domain) {
        msg = g_strdup_printf("%s: %s", log_domain, message);
    } else {
        msg = g_strdup(message);
    }
    dialog_show(window, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
    g_free(msg);
}

G_MODULE_EXPORT gboolean
window_state_changed(GtkWidget *caller, GdkEvent *event, gpointer data)
{
    ConfigGUIOpts opts = config_gui_get_opts();

#if GTK_CHECK_VERSION(2,10,0)
    if (!opts.show_tray_icon || !tray_icon_get_is_shown()) {
        return FALSE;
    }

    if (    (event->window_state.new_window_state & GDK_WINDOW_STATE_ICONIFIED) == GDK_WINDOW_STATE_ICONIFIED &&
            (event->window_state.changed_mask & GDK_WINDOW_STATE_ICONIFIED) == GDK_WINDOW_STATE_ICONIFIED) {
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(caller), TRUE);
    }
#endif

    return FALSE;
}

static gboolean
initialisation(gpointer data)
{
    GtkWidget *windowMain = get_widget("windowMain");
    char *errmsg = NULL;
#ifdef NNTPGRAB_NETWORKED
    GtkWidget *cmbConnectHostname;
    GtkWidget *spinConnectPort;
    GtkWidget *windowConnect;
    char *hostname;
    int port;
#else
    char *warnings = NULL;
#endif

    gtk_widget_show(windowMain);

#ifdef NNTPGRAB_NETWORKED
    cmbConnectHostname = get_widget("cmbConnectHostname");
    spinConnectPort = get_widget("spinConnectPort");
    windowConnect = get_widget("windowConnect");
    hostname = gtk_combo_box_get_active_text(GTK_COMBO_BOX(cmbConnectHostname));
    port = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinConnectPort));

    if (!(core = nntpgrab_networked_init())) {
        g_free(hostname);
        g_critical(_("NNTPGrab Initialisation FAILED\n"));
        gtk_main_quit();
        return FALSE;
    }

    if (!nntpgrab_networked_connect(core, hostname, port, NULL, NULL, FALSE, &errmsg)) {
        g_free(hostname);
        g_critical(_("Unable to connect to NNTPGrab server:\n%s\n"), errmsg);
        g_free(errmsg);
        gtk_main_quit();
        return FALSE;
    }

    g_free(hostname);

    gtk_widget_hide(windowConnect);

#else

    if (!(core = nntpgrab_core_init(NNTPGRAB_API_VERSION, &errmsg, &warnings))) {
        g_critical(_("NNTPGrab Initialisation FAILED:\n%s\n"), errmsg);
        g_free(errmsg);
        gtk_main_quit();
        return FALSE;
    }

    if (warnings) {
        GtkWidget *windowMain = get_widget("windowMain");
        dialog_show(windowMain, warnings, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
        g_free(warnings);
    }
#endif

    debug_initialize();
    callbacks_initialize();
    info_initialize();
    config_initialize();
    config_display_initialize();
#ifndef DARWIN
    nzb_initialize();
#endif
    connections_initialize();
    queue_initialize();
    notify_initialize();
    monitor_initialize();
    par2_initialize();

    g_signal_connect(G_OBJECT(windowMain), "window_state_event",
                           G_CALLBACK(window_state_changed), NULL);

    // Did we receive a NZB file on the command line ?
    if (nzb_file) {
        do_open_nzb_file(nzb_file);
        nzb_file = NULL;
    }

    return FALSE;
}

G_MODULE_EXPORT void
on_btnConnectToServer_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowConnect = get_widget("windowConnect");

    // Save the hostname into the GUI Configuration
    ConfigGUIOpts opts = config_gui_get_opts();
    GtkWidget *cmbConnectHostname = get_widget("cmbConnectHostname");
    char *hostname = gtk_combo_box_get_active_text(GTK_COMBO_BOX(cmbConnectHostname));

    // Have we seen this hostname before ?
    if (!strcmp(hostname, opts.hostname1)) {
        // No need to do anything, item is already at the top of the list
    } else if (!strcmp(hostname, opts.hostname2)) {
        // Swap positions
        strncpy(opts.hostname2, opts.hostname1, sizeof(opts.hostname2));
        strncpy(opts.hostname1, hostname, sizeof(opts.hostname1));
    } else if (!strcmp(hostname, opts.hostname3)) {
        // Push items 1 and 2 one place down and set the first item to the hostname
        strncpy(opts.hostname3, opts.hostname2, sizeof(opts.hostname3));
        strncpy(opts.hostname2, opts.hostname1, sizeof(opts.hostname2));
        strncpy(opts.hostname1, hostname, sizeof(opts.hostname1));
    } else if (!strcmp(hostname, opts.hostname4)) {
        strncpy(opts.hostname4, opts.hostname3, sizeof(opts.hostname4));
        strncpy(opts.hostname3, opts.hostname2, sizeof(opts.hostname3));
        strncpy(opts.hostname2, opts.hostname1, sizeof(opts.hostname2));
        strncpy(opts.hostname1, hostname, sizeof(opts.hostname1));
    } else if (!strcmp(hostname, opts.hostname5)) {
        strncpy(opts.hostname5, opts.hostname4, sizeof(opts.hostname5));
        strncpy(opts.hostname4, opts.hostname3, sizeof(opts.hostname4));
        strncpy(opts.hostname3, opts.hostname2, sizeof(opts.hostname3));
        strncpy(opts.hostname2, opts.hostname1, sizeof(opts.hostname2));
        strncpy(opts.hostname1, hostname, sizeof(opts.hostname1));
    } else {
        // Hostname not seen at all, move all the items one place down (forget
        // about the last item) and put the hostname in front
        strncpy(opts.hostname5, opts.hostname4, sizeof(opts.hostname5));
        strncpy(opts.hostname4, opts.hostname3, sizeof(opts.hostname4));
        strncpy(opts.hostname3, opts.hostname2, sizeof(opts.hostname3));
        strncpy(opts.hostname2, opts.hostname1, sizeof(opts.hostname2));
        strncpy(opts.hostname1, hostname, sizeof(opts.hostname1));
    }

    config_gui_set_opts(opts);

    gtk_widget_set_sensitive(windowConnect, FALSE);
    g_idle_add(initialisation, NULL);
}

G_MODULE_EXPORT void
on_btnConnectQuit_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowConnect = get_widget("windowConnect");

    gtk_widget_hide(windowConnect);

    gtk_main_quit();
}

#ifdef _MSC_VER
void AppInvalidParameterHandler(const wchar_t* expression, const wchar_t* function, const wchar_t* file, unsigned int line,	uintptr_t pReserved ) {
	fprintf(
		stderr,
		"Invalid parameter detected in function %s. File: %s Line: %d\n",
		function,
		file,
		line
	);
	fprintf(
		stderr,
		"Expression: %s\n",
		expression
	);
	// Cause a Debug Breakpoint.
	DebugBreak();
}
#endif

int
main(int argc, char *argv[])
{
    char *filename;
    GError *err = NULL;
    GtkWidget *windowMain;
    GtkWidget *btnDonate;
    GtkWidget *icon;
    GdkPixbuf *pixbuf;
    const char *path;
    int i;
#ifdef NNTPGRAB_NETWORKED
    GtkWidget *windowConnect;
    GtkWidget *cmbConnectHostname;
    GtkListStore *store;
    ConfigGUIOpts opts;
    GtkTreeIter iter;
#endif
#ifdef WIN32
    char *dirname;
    char *dirname2;
    WORD wVersionRequested;
    WSADATA wsaData;
    int ret;
#endif
#if GTK_CHECK_VERSION(2,12,0)
    const char *interface_file = "nntpgrab_gui.ui";
#else
    const char *interface_file = "nntpgrab_gui.glade";
#endif

#ifdef WIN32
#ifdef _MSC_VER
    _set_invalid_parameter_handler(AppInvalidParameterHandler);
#else
    // exchndl.dll is part of Dr. MinGW. By loading this DLL, every time the
    // program crashes a .RPT file will be created containing the backtrace
    LoadLibraryA("exchndl.dll");
#endif
#endif

    g_set_application_name("NNTPGrab");
    g_set_prgname("NNTPGrab");
    gtk_init(&argc, &argv);

#ifdef WIN32
    dirname = g_win32_get_package_installation_directory_of_module(NULL); 
    dirname2 = g_build_path(G_DIR_SEPARATOR_S, dirname, "\\share\\locale", NULL);
    bindtextdomain (GETTEXT_PACKAGE, dirname2);
    g_free(dirname2);
#else
    bindtextdomain (GETTEXT_PACKAGE, NNTPGRABLOCALEDIR);
#endif
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    g_log_set_handler(NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);

#ifdef WIN32
    // Initialise Winsock
    wVersionRequested = MAKEWORD(2, 2);

    ret = WSAStartup(wVersionRequested, &wsaData);
    if (ret != 0) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        g_error(_("Winsock could not be initialised"));
        return -1;
    }

    /* Confirm that the WinSock DLL supports 2.2.        */
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */

    if (LOBYTE( wsaData.wVersion ) != 2 ||
        HIBYTE( wsaData.wVersion ) != 2) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup();

        g_error(_("Winsock 2.2 or higher is required for this program"));

        return -1;
    }
#endif

    if (argc > 1) {
        // We have an extra argument! Probably the location of a NZB file
        // Is one our arguments really a filename ?
        for (i = 1; i < argc; i++) {
            char *filename;

#ifdef WIN32
            filename = g_convert(argv[i], -1, "utf-8", "windows-1252", NULL, NULL, NULL);
#else
            filename = g_strdup(argv[i]);
#endif
            if (g_file_test(filename, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
                nzb_file = g_strdup(filename);
            }
            g_free(filename);
        }

        if (nzb_file) {
            // Is there already another NNTPGrab instance running ?
            if (ipc_test_is_nntpgrab_already_running()) {
                // Send a message to that other process and commit suicide
                ipc_send_open_nzb_file_request(nzb_file);
                return(1);
            }

            // Nope, open the NZB Import dialog as soon as we're ready to go
            // This will be done automatically due to the fact that nzb_file is a global variable
        }
    }

    /* load the interface */
#ifdef WIN32
    path = g_build_path(G_DIR_SEPARATOR_S, dirname, "\\share\\nntpgrab", NULL);
    g_free(dirname);
#else
    if (g_getenv("NNTPGRAB_DATA_DIR")) {
        path = g_getenv("NNTPGRAB_DATA_DIR");
    } else {
        path = SHARE_DIR;
    }
#endif

    filename = g_build_filename(path, G_DIR_SEPARATOR_S, interface_file, NULL);

#if GTK_CHECK_VERSION(2,12,0)
    // Register GtkHTML with GObject so that GtkBuilder can use it
    gtk_html_get_type();

    builder = gtk_builder_new();
    if (!gtk_builder_add_from_file(builder, filename, &err)) {
        builder = NULL;
        g_error(err->message);
    }

    gtk_builder_connect_signals(builder, NULL);

    // GtkBuilder doesn't support a custom creation function yet, so the GtkHTML
    // component needs to be manually configured
    initialize_html_widget();
#else
    xml = glade_xml_new(filename, NULL, NULL);

    /* connect the signals in the interface */
    glade_xml_signal_autoconnect(xml);
#endif

    g_free(filename);

#ifdef DARWIN
    ige_mac_menu_set_menu_bar(GTK_MENU_SHELL(get_widget("menubar1")));
    gtk_widget_hide(get_widget("menubar1"));
#endif

    // Set the icon of the window
    windowMain = get_widget("windowMain");
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) &icon_64x64_xpm);
    gtk_window_set_icon(GTK_WINDOW(windowMain), pixbuf);
    g_object_unref(pixbuf);

    // Set the donate icon of the window
    btnDonate = get_widget("btnDonate");
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) &donate_xpm);
    icon = gtk_image_new_from_pixbuf(pixbuf);
    gtk_widget_show(icon);
    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON(btnDonate), icon);
    g_object_unref(pixbuf);

#ifdef DARWIN
    // Due to the way file associations work on Mac OS X, we need to make
    // sure the NZB dialog can be used before the file association is set up
    nzb_initialize();
#endif

    if (!ipc_initialize()) {
        // There is already another NNTPGrab process active!
        return 1;
    }
	
    // We need to have the GUI configuration loaded at this point because the
    // Network GUI requires data from it
    config_gui_initialize();

#ifdef NNTPGRAB_NETWORKED
    // Show the connect dialog so the user can enter where the NNTPGrab server is running
    windowConnect = get_widget("windowConnect");
    cmbConnectHostname = get_widget("cmbConnectHostname");

    gtk_widget_show(windowConnect);

    store = gtk_list_store_new(1, G_TYPE_STRING);
    gtk_combo_box_set_model(GTK_COMBO_BOX(cmbConnectHostname), GTK_TREE_MODEL(store));
    gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(cmbConnectHostname), 0);

    opts = config_gui_get_opts();

    if (strlen(opts.hostname1) > 0) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, opts.hostname1, -1);
    }

    if (strlen(opts.hostname2) > 0) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, opts.hostname2, -1);
    }

    if (strlen(opts.hostname3) > 0) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, opts.hostname3, -1);
    }

    if (strlen(opts.hostname4) > 0) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, opts.hostname4, -1);
    }

    if (strlen(opts.hostname5) > 0) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, opts.hostname5, -1);
    }

    gtk_combo_box_set_active(GTK_COMBO_BOX(cmbConnectHostname), 0);
#else
    // Perform the main initialisation
    g_idle_add(initialisation, NULL);
#endif

    /* start the event loop */
    gtk_main();

#if GTK_CHECK_VERSION(2,10,0)
    if (tray_icon_get_is_shown()) {
        tray_destroy();
    }
#endif

#ifdef NNTPGRAB_NETWORKED
    if (core) {
        nntpgrab_networked_cleanup(core);
    }
#else
    nntpgrab_core_cleanup(core);
#endif

    return 0;
}

#ifdef WIN32
static HINSTANCE instance = NULL;

HINSTANCE
get_hinstance(void)
{
    return instance;
}

int _stdcall
WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, char *lpszCmdLine, int nCmdShow)
{
    instance = hInstance;
    return main(__argc, __argv);
}

#endif
