/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NNTPGRAB_PLUGIN_SCHEDULAR_H_
#define _NNTPGRAB_PLUGIN_SCHEDULAR_H_

#include "nntpgrab_plugin_base.h"
#include "collections.h"

// These are all the functions that the plugin can call to communicate with the NNTPGrab core
typedef struct _plugin_schedular_imported_funcs {
    int version;

    void (*task_added) (const NNTPCollection *collection, const NNTPFile *file);
    void (*task_removed) (const char *collection_name, const char *subject, guint64 total_size, guint64 total_size_remaining);
    void (*task_state_changed) (const char *collection_name, const char *subject, TaskState old_state, TaskState new_state);
    void (*collection_added) (const char *collection_name);
    void (*collection_removed) (const char *collection_name);

    void (*notify_schedular_state_changed) (SchedularState new_state, const char *reason);

    GList * (*config_get_avail_servers) (void);
    ConfigServer * (*config_get_server_info) (const char *servername);
    ConfigOpts (*config_get_opts) (void);

    gpointer (*nntp_connection_initialize) (const char *servername, NNTPGrabErrCode *err);
    void (*nntp_connection_destroy) (gpointer obj);
    NNTPGrabErrCode (*nntp_connection_get_part) (gpointer obj, const char *collection_name, const NNTPFile *file, const NNTPPart *part, char **reason);

    gboolean (*decode_file) (const char *collection_name, const NNTPFile *file, int *saved_errno);

    void (*emit_fatal_error) (const char *errmsg);
    void (*emit_part_download_complete) (const char *servername, gpointer connection, const char *collection_name, const char *subject, int partnum, int size);
    void (*emit_part_download_failed) (const char *servername, gpointer connection, const char *collection_name, const char *subject, int partnum, int size, gboolean all_servers_tried);
    void (*emit_task_moved) (const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position);
    void (*emit_collection_moved) (const char *collection_name, int old_position, int new_position);
    void (*emit_file_download_state_update) (const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining);
} PluginSchedularImportedFuncs;

// The plugin needs to exported these methods :

/**
 * Initialize the plugin
 *
 * @arg     funcs   A structure containing all the callback functions the plugin can call to notify the NNTPGrab Core
 * @arg     errmsg  Pointer to a g_malloc'ed error message which will be set when the function returns FALSE
 *
 * @return          TRUE when the initialization was successfull
 */
gboolean     nntpgrab_plugin_schedular_initialize(PluginSchedularImportedFuncs funcs, char **errmsg);

/**
 * Destroy the plugin
 */
void         nntpgrab_plugin_schedular_destroy(void);

/**
 * Return the API version of the plugin. This is used to verify that the plugin is compatible
 * with the interface the NNTPGrab Core expects
 *
 * @return  NNTPGRAB_PLUGIN_API_VERSION
 */
int          nntpgrab_plugin_schedular_get_version(void);

/**
 * Start the schedular thread. Any items in the download queue will be downloaded
 *
 * @return  FALSE when the queue is already started
 */
gboolean     nntpgrab_plugin_schedular_start(void);

/**
 * Stop/pause the schedular thread. Active download jobs will be finished, but no new ones will be started.
 *
 * @arg     The reason why the schedular needs to be stopped. Needs to be NULL for external uses
 *
 * @return  FALSE when the queue is already stopped
 */
gboolean    nntpgrab_plugin_schedular_stop(const char *reason);

/**
 * Retrieves the status of the schedular
 *
 * @return  SCHEDULAR_STATE_RUNNING when the schedular is running,
 *          SCHEDULAR_STATE_STOPPING when the schedular is waiting for jobs to finish so the schedular can be stopped
 *          SCHEDULAR_STATE_STOPPED when the schedular is not running
 */
SchedularState   nntpgrab_plugin_schedular_get_state(void);

/**
 * Add a new task to the download queue
 *
 * @arg collection_name     The name of the collection where this file belongs to
 * @arg file                The structure containing what data needs to be downloaded
 * @arg errmsg              Pointer to a g_malloc'ed error message which will be set when the function returns FALSE
 *
 */
gboolean     nntpgrab_plugin_schedular_add_task_to_queue(const char *collection_name, NNTPFile *file, char **errmsg);

/**
 * Remove a task from the download queue
 *
 * @arg collection_name     The name of the collection where the task belongs to
 * @arg subject             The subject of the task which needs to be removed
 * @arg errmsg              Pointer to a g_malloc'ed error message which will be set when the function returns FALSE
 *
 * @return                  TRUE on success, FALSE when the subject was not found in the given collection
 */
gboolean     nntpgrab_plugin_schedular_del_task_from_queue(const char *collection_name, const char *subject, char **errmsg);

/**
 * Restarts a task. All parts will be downloaded and decoded again
 *
 * @arg collection_name     The name of the collection where the task belongs to
 * @arg subject             The subject of the task which needs to be restarted
 * @arg errmsg              Pointer to a g_malloc'ed error message which will be set when the function returns FALSE
 *
 * @return                  TRUE on success, FALSE when the subject was not found in the given collection
 */
gboolean     nntpgrab_plugin_schedular_restart_task(const char *collection_name, const char *subject, char **errmsg);

/**
 * Saves the download queue to disk
 *
 * @arg errmsg              Pointer to a g_malloc'ed error message which will be set when the function returns FALSE
 *
 */
gboolean     nntpgrab_plugin_schedular_save_queue(char **errmsg);

/**
 * Get a list of all the tasks which are already in the download queue
 *
 * @arg foreach_func        Function which needs to be called with each collection currently in the queue
 * @arg data                user-specified data
 */
void         nntpgrab_plugin_schedular_foreach_task(GFunc foreach_func, gpointer data);

/**
 * Moves one item in the download queue to a new position and collection
 *
 * @arg collection_name_src     Original collection name
 * @arg subject_src             The name of the subject which needs to be moved
 * @arg collection_name_dest    The new location of the subject
 * @arg position_dest           The new position of the item in the collection collection_name_dest
 *
 * @return  TRUE when the task has been successfully moved
 */
gboolean
nntpgrab_plugin_schedular_move_task(const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest);

/**
 * Moves the position of one collection in the download queue
 *
 * @arg collection_name         Collection name
 * @arg new_position            The new position of the collection
 *
 * @return  TRUE when the collection has been successfully moved
 */
gboolean
nntpgrab_plugin_schedular_move_collection(const char *collection_name, int new_position);

#endif /* _NNTPGRAB_PLUGIN_SCHEDULAR_H_ */
