/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <glib-object.h>
#include <glib/gi18n.h>

#include "nntpgrab_plugin_schedular.h"
#include "schedular_plugin.h"

void
decoder_thread_func(gpointer data, gpointer user_data)
{
    DecoderData *decoder_data = (DecoderData *) data;
    int saved_errno = 0;
    static time_t last_timestamp = 0;

    imported_funcs_task_state_changed(decoder_data->collection_name, decoder_data->file->subject, decoder_data->file->status, TASK_STATE_DECODING);

    g_static_mutex_lock(&decoder_data->schedular->mutex);
    decoder_data->file->status = TASK_STATE_DECODING;
    g_static_mutex_unlock(&decoder_data->schedular->mutex);

    // Perform the actual decoding
    if (!imported_funcs_decode_file(decoder_data->collection_name, decoder_data->file, &saved_errno)) {
        if (saved_errno == ENOSPC) {
            // No free space available, stop the schedular
            stop_schedular_from_seperate_thread(strerror(saved_errno));

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_mutex_lock(&decoder_data->schedular->mutex);
            decoder_data->file->now_decoding = FALSE;
            imported_funcs_task_state_changed(decoder_data->collection_name, decoder_data->file->subject, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->schedular->mutex);

            g_free(decoder_data);

            return;
        } else if (saved_errno == ENOENT) {
            // No parts available, ignore
        } else if (saved_errno == EEXIST) {
            // File already exists, ignore
        } else {
            // Unknown decode error occured, stop the schedular
            char *msg = g_strdup_printf("Unknown decode error occured, %s (%i)", strerror(saved_errno), saved_errno);
            stop_schedular_from_seperate_thread(msg);
            g_free(msg);

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_mutex_lock(&decoder_data->schedular->mutex);
            decoder_data->file->now_decoding = FALSE;
            imported_funcs_task_state_changed(decoder_data->collection_name, decoder_data->file->subject, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->schedular->mutex);

            g_free(decoder_data);

            return;
        }
    }

    // Mark the file as completed
    g_static_mutex_lock(&decoder_data->schedular->mutex);

    // Was the task restarted while the decoding was busy?
    if (decoder_data->file->now_decoding) {
        // Nope, mark the file decode as completed
        decoder_data->file->file_is_downloaded = TRUE;
        decoder_data->file->now_decoding = FALSE;

        imported_funcs_task_state_changed(decoder_data->collection_name, decoder_data->file->subject, decoder_data->file->status, TASK_STATE_FINISHED);
        decoder_data->file->status = TASK_STATE_FINISHED;
    }

    g_static_mutex_unlock(&decoder_data->schedular->mutex);

    // Is this file a PAR2 file?

    g_free(decoder_data);

    // Save the download queue only when there has been some time between now and the previous time we've decoded something
    if (time(NULL) > last_timestamp + 60) {
        schedular_plugin_save_queue(NULL);
        last_timestamp = time(NULL);
    }
}

