/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include <glib.h>
#include <glib-object.h>

#include "nntpgrab.h"

static void
config_changed (NNTPGrabCore *core, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup("CONFIG_CHANGED\n"));
}

static void
part_download_start (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_DOWNLOAD_START %s\t%i\t%s\t%s\t%i\n", servername, conn_id, collection_name, subject, part_num));
}

static void
part_done (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_DONE %s\t%i\t%s\t%s\t%i\t%i\n", servername, conn_id, collection_name, subject, part_num, size));
}

static void
part_failed (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    const char *all_servers_tried_str;

    if (all_servers_tried) {
        all_servers_tried_str = "TRUE";
    } else {
        all_servers_tried_str = "FALSE";
    }

    g_async_queue_push(queue, g_strdup_printf("PART_FAILED %s\t%i\t%s\t%s\t%i\t%i\t%s\n", servername, conn_id, collection_name, subject, part_num, size, all_servers_tried_str));
}

static void
traffic_monitor_update (NNTPGrabCore *core, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 stamp, double average, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

#if SIZEOF_LONG == 8
    g_async_queue_push(queue, g_strdup_printf("TRAFFIC_MONITOR_UPDATE %i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%li\t%.2f\n", bytes_received1, bytes_received2, bytes_received3, bytes_received4, bytes_received5, bytes_received6, bytes_received7, bytes_received8, bytes_received9, bytes_received10, stamp, average));
#else /* SIZEOF_LONG == 4 */
    g_async_queue_push(queue, g_strdup_printf("TRAFFIC_MONITOR_UPDATE %i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%lli\t%.2f\n", bytes_received1, bytes_received2, bytes_received3, bytes_received4, bytes_received5, bytes_received6, bytes_received7, bytes_received8, bytes_received9, bytes_received10, stamp, average));
#endif
}

static void
part_progress_update (NNTPGrabCore *core, const char *servername, int conn_id, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("PART_PROGRESS_UPDATE %s\t%i\t%s\t%i\t%i\t%i\n", servername, conn_id, subject, part_num, bytes_downloaded, bytes_total));
}

static void
collection_added (NNTPGrabCore *core, const char *collection_name, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_ADDED %s\n", collection_name));
}

static void
collection_removed (NNTPGrabCore *core, const char *collection_name, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_REMOVED %s\n", collection_name));
}

static void
file_added (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_size, guint64 total_size_remaining, TaskState status, int num_parts, GList *groups, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *groups_str = NULL;
    GList *list;
    char *status_str;

    list = groups;
    while (list) {
        char *tmp;
        const char *groupname = list->data;

        if (groups_str) {
            tmp = g_strdup_printf("%s|%s", groups_str, groupname);
            g_free(groups_str);
            groups_str = tmp;
        } else {
            groups_str = g_strdup(groupname);
        }

        list = g_list_next(list);
    }

    switch (status) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            status_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            status_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            status_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            status_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            status_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED:
        default:
            status_str = "FINISHED";
            break;
    };

#if SIZEOF_LONG == 8
    g_async_queue_push(queue, g_strdup_printf("FILE_ADDED %s\t%s\t%s\t%li\t%li\t%li\t%li%s\t%i\t%s\n", collection_name, subject, poster, stamp, file_size, total_size, total_size_remaining, status_str, num_parts, groups_str));
#else /* SIZEOF_LONG == 4 */
    g_async_queue_push(queue, g_strdup_printf("FILE_ADDED %s\t%s\t%s\t%lli\t%lli\t%lli\t%lli\t%s\t%i\t%s\n", collection_name, subject, poster, stamp, file_size, total_size, total_size_remaining, status_str, num_parts, groups_str));
#endif
    g_free(groups_str);
}

static void
file_removed (NNTPGrabCore *core, const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

#if SIZEOF_LONG == 8
    g_async_queue_push(queue, g_strdup_printf("FILE_REMOVED %s\t%s\t%li\t%li\n", collection_name, subject, total_file_size, total_file_size_remaining));
#else /* SIZEOF_LONG == 4 */
    g_async_queue_push(queue, g_strdup_printf("FILE_REMOVED %s\t%s\t%lli\t%lli\n", collection_name, subject, total_file_size, total_file_size_remaining));
#endif
}

static void
file_download_state_update (NNTPGrabCore *core, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

#if SIZEOF_LONG == 8
    g_async_queue_push(queue, g_strdup_printf("FILE_DOWNLOAD_STATE_UPDATE %s\t%s\t%i\t%i\t%i\t%li\t%li\t%li\t%li\n", collection_name, subject, num_parts_total, num_parts_done, num_parts_failed, file_size, file_size_remaining, total_file_size, total_file_size_remaining));
#else /* SIZEOF_LONG == 4 */
    g_async_queue_push(queue, g_strdup_printf("FILE_DOWNLOAD_STATE_UPDATE %s\t%s\t%i\t%i\t%i\t%lli\t%lli\t%lli\t%lli\n", collection_name, subject, num_parts_total, num_parts_done, num_parts_failed, file_size, file_size_remaining, total_file_size, total_file_size_remaining));
#endif
}

static void
file_state_changed (NNTPGrabCore *core, const char *collection_name, const char *subject, TaskState old_state, TaskState new_state, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *old_state_str;
    char *new_state_str;

    switch(old_state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            old_state_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            old_state_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            old_state_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            old_state_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            old_state_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED:
        default:
            old_state_str = "FINISHED";
            break;
    };

    switch(new_state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            new_state_str = "WAITING_FOR_DOWNLOAD";
            break;

        case TASK_STATE_DOWNLOADING:
            new_state_str = "DOWNLOADING";
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            new_state_str = "WAITING_FOR_DECODE";
            break;

        case TASK_STATE_DECODING:
            new_state_str = "DECODING";
            break;

        case TASK_STATE_SKIPPED:
            new_state_str = "SKIPPED";
            break;

        case TASK_STATE_FINISHED:
        default:
            new_state_str = "FINISHED";
            break;
    };

    g_async_queue_push(queue, g_strdup_printf("FILE_STATE_CHANGED %s\t%s\t%s\t%s\n", collection_name, subject, old_state_str, new_state_str));
}

static void
connection_connecting (NNTPGrabCore *core, const char *servername, int conn_id, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_CONNECTING %s\t%i\n", servername, conn_id));
}

static void
connection_connected (NNTPGrabCore *core, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_CONNECTED %s\t%i\t%s\n", servername, conn_id, welcome_msg));
}

static void
connection_disconnect (NNTPGrabCore *core, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *disconnect_type_str;

    switch (disconnect_type) {
        case DISCONNECT_NORMAL:
            disconnect_type_str = "NORMAL";
            break;

        case DISCONNECT_NO_SUCH_HOST:
            disconnect_type_str = "NO_SUCH_HOST";
            break;

        case DISCONNECT_CONNECTION_REFUSED:
            disconnect_type_str = "CONNECTION_REFUSED";
            break;

        case DISCONNECT_TOO_MANY_CONNECTIONS:
            disconnect_type_str = "TOO_MANY_CONNECTIONS";
            break;

        case DISCONNECT_CONNECT_TIMEOUT:
            disconnect_type_str = "CONNECT_TIMEOUT";
            break;

        case DISCONNECT_READ_ERROR:
            disconnect_type_str = "READ_ERROR";
            break;

        case DISCONNECT_READ_TIMEOUT:
            disconnect_type_str = "READ_TIMEOUT";
            break;

        case DISCONNECT_WRITE_ERROR:
            disconnect_type_str = "WRITE_ERROR";
            break;

        case DISCONNECT_IDLE_TIMEOUT:
            disconnect_type_str = "IDLE_TIMEOUT";
            break;

        case DISCONNECT_INVALID_MSG:
            disconnect_type_str = "INVALID_MSG";
            break;

        case DISCONNECT_LOGIN_FAILURE:
            disconnect_type_str = "LOGIN_FAILURE";
            break;

        case DISCONNECT_UNEXPECTED:
        default:
            disconnect_type_str = "UNEXPECTED";
            break;
    };

    g_async_queue_push(queue, g_strdup_printf("CONNECTION_DISCONNECT %s\t%i\t%s\t%s\n", servername, conn_id, disconnect_type_str, (msg) ? msg : ""));
}

static void
schedular_state_changed (NNTPGrabCore *core, SchedularState new_state, const char *reason, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *reason2 = NULL;

    // strip out any newlines
    if (reason) {
        char *ptr;

        reason2 = g_strdup(reason);
        ptr = reason2;

        while (*ptr) {
            if (*ptr == '\n') {
                *ptr = ' ';
            }
            ptr++;
        }
    }

    switch (new_state) {
        case SCHEDULAR_STATE_RUNNING:
            g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED RUNNING\t\n"));
            break;

        case SCHEDULAR_STATE_STOPPING:
            if (reason2) {
                g_async_queue_push(queue, g_strdup_printf("SCHEDULAR_STATE_CHANGED STOPPING\t%s\n", reason2));
            } else {
                g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED STOPPING\t\n"));
            }

            break;

        case SCHEDULAR_STATE_STOPPED:
        default:
            g_async_queue_push(queue, g_strdup("SCHEDULAR_STATE_CHANGED STOPPED\t\n"));
            break;
    };

    if (reason2) {
        g_free(reason2);
    }
}

static void
fatal_error (NNTPGrabCore *core, const char *errmsg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(errmsg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("FATAL_ERROR %s\n", msg));

    g_free(msg);
}

static void
warning_message (NNTPGrabCore *core, const char *warn_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(warn_msg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("WARNING_MESSAGE %s\n", msg));

    g_free(msg);
}

static void
debug_message (NNTPGrabCore *core, GThread *thread, const char *debug_msg, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;
    char *msg = g_strdup(debug_msg);
    char *ptr = msg;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    g_async_queue_push(queue, g_strdup_printf("DEBUG_MESSAGE %s\n", msg));

    g_free(msg);
}

static void
task_moved (NNTPGrabCore *core, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("TASK_MOVED %s\t%s\t%s\t%i\t%i\n", orig_collection_name, subject, new_collection_name, old_position, new_position));
}

static void
collection_moved (NNTPGrabCore *core, const char *collection_name, int old_position, int new_position, gpointer data)
{
    GAsyncQueue *queue = (GAsyncQueue*) data;

    g_async_queue_push(queue, g_strdup_printf("COLLECTION_MOVED %s\t%i\t%i\n", collection_name, old_position, new_position));
}

void
connect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *queue)
{
    gpointer data = queue;

    g_signal_connect(core, "config_changed", G_CALLBACK(config_changed), data);
    g_signal_connect(core, "part_download_start", G_CALLBACK(part_download_start), data);
    g_signal_connect(core, "part_done", G_CALLBACK(part_done), data);
    g_signal_connect(core, "part_failed", G_CALLBACK(part_failed), data);
    g_signal_connect(core, "traffic_monitor_update", G_CALLBACK(traffic_monitor_update), data);
    g_signal_connect(core, "part_progress_update", G_CALLBACK(part_progress_update), data);
    g_signal_connect(core, "collection_added", G_CALLBACK(collection_added), data);
    g_signal_connect(core, "collection_removed", G_CALLBACK(collection_removed), data);
    g_signal_connect(core, "file_added", G_CALLBACK(file_added), data);
    g_signal_connect(core, "file_removed", G_CALLBACK(file_removed), data);
    g_signal_connect(core, "file_download_state_update", G_CALLBACK(file_download_state_update), data);
    g_signal_connect(core, "file_state_changed", G_CALLBACK(file_state_changed), data);
    g_signal_connect(core, "connection_connecting", G_CALLBACK(connection_connecting), data);
    g_signal_connect(core, "connection_connected", G_CALLBACK(connection_connected), data);
    g_signal_connect(core, "connection_disconnect", G_CALLBACK(connection_disconnect), data);
    g_signal_connect(core, "schedular_state_changed", G_CALLBACK(schedular_state_changed), data);
    g_signal_connect(core, "fatal_error", G_CALLBACK(fatal_error), data);
    g_signal_connect(core, "warning_message", G_CALLBACK(warning_message), data);
    g_signal_connect(core, "debug_message", G_CALLBACK(debug_message), data);
    g_signal_connect(core, "task_moved", G_CALLBACK(task_moved), data);
    g_signal_connect(core, "collection_moved", G_CALLBACK(collection_moved), data);
}

void
disconnect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *queue)
{
    gpointer data = queue;

    g_signal_handlers_disconnect_by_func(core, config_changed, data);
    g_signal_handlers_disconnect_by_func(core, part_download_start, data);
    g_signal_handlers_disconnect_by_func(core, part_done, data);
    g_signal_handlers_disconnect_by_func(core, part_failed, data);
    g_signal_handlers_disconnect_by_func(core, traffic_monitor_update, data);
    g_signal_handlers_disconnect_by_func(core, part_progress_update, data);
    g_signal_handlers_disconnect_by_func(core, collection_added, data);
    g_signal_handlers_disconnect_by_func(core, collection_removed, data);
    g_signal_handlers_disconnect_by_func(core, file_added, data);
    g_signal_handlers_disconnect_by_func(core, file_removed, data);
    g_signal_handlers_disconnect_by_func(core, file_download_state_update, data);
    g_signal_handlers_disconnect_by_func(core, file_state_changed, data);
    g_signal_handlers_disconnect_by_func(core, connection_connecting, data);
    g_signal_handlers_disconnect_by_func(core, connection_connected, data);
    g_signal_handlers_disconnect_by_func(core, connection_disconnect, data);
    g_signal_handlers_disconnect_by_func(core, schedular_state_changed, data);
    g_signal_handlers_disconnect_by_func(core, fatal_error, data);
    g_signal_handlers_disconnect_by_func(core, warning_message, data);
    g_signal_handlers_disconnect_by_func(core, debug_message, data);
    g_signal_handlers_disconnect_by_func(core, task_moved, data);
    g_signal_handlers_disconnect_by_func(core, collection_moved, data);
}
