/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>

#ifndef _MSC_VER
#include <strings.h>
#endif

#include <glib/gi18n.h>
#include "nntpgrab_utils.h"
#include "nntpgrab_types.h"

gboolean
nntpgrab_utils_strip_subject(const char *subject, char **suffix, int *file_num, int *total_files, char **filename, char **extension, NNTPFileType *file_type, int *par2_startnum, int *num_par2_blocks, int *part_num, int *total_parts)
{
    const char **values;
    static NNTPGRAB_REGEX *re_filename_and_extension = NULL;
    NNTPFileType type;

    if (!re_filename_and_extension) {
        re_filename_and_extension = nntpgrab_utils_regex_compile("(?:.*[\"'\\]]|[^ ]+ |^)([^\"'\\]]+\\.([fF][lL][aA][cC]|[pP][aA][rR]2|[a-zA-Z0-9]{3}|[tT][sS]))(?:[\"'\\] ].*)\\(([0-9]+)/([0-9]+)\\)$");
    }

    values = nntpgrab_utils_regex_match(re_filename_and_extension, subject);
    if (values == NULL) {
        return FALSE;
    }

    if (filename) {
        *filename = g_strdup(values[1]);
    }

    if (extension) {
        *extension = g_strdup(values[2]);
    }

    if (file_type || par2_startnum || num_par2_blocks) {
        if (!g_strcasecmp(values[2], "rar")) {
            type = NNTP_FILE_TYPE_RAR;
        } else if (!g_strcasecmp(values[2], "par2")) {
            type = NNTP_FILE_TYPE_PAR2;
        } else if (!g_strcasecmp(values[2], "par")) {
            type = NNTP_FILE_TYPE_PAR;
        } else if (!g_strcasecmp(values[2], "sfv")) {
            type = NNTP_FILE_TYPE_SFV;
        } else if (!g_strcasecmp(values[2], "nfo")) {
            type = NNTP_FILE_TYPE_NFO;
        } else if (!g_strcasecmp(values[2], "nzb")) {
            type = NNTP_FILE_TYPE_NZB;
        } else {
            type = NNTP_FILE_TYPE_OTHER;
        }

        if (file_type) {
            *file_type = type;
        }

        if (par2_startnum || num_par2_blocks) {
            *par2_startnum = -1;
            *num_par2_blocks = -1;

            if (type == NNTP_FILE_TYPE_PAR2) {
                // Find out the block numbers of this PAR2 file.
                // FIXME: This should be integrated in the main regexp
                static NNTPGRAB_REGEX *re_par2_blocknumbers = NULL;
                const char **values_par2;

                if (!re_par2_blocknumbers) {
                    re_par2_blocknumbers = nntpgrab_utils_regex_compile(".*\\.[vV][oO][lL](\\d+)\\+(\\d+)\\.[pP][aA][rR]2$");
                }

                values_par2 = nntpgrab_utils_regex_match(re_par2_blocknumbers, values[1]);
                if (values_par2 && values_par2[0] && values_par2[1] && values_par2[2]) {
                    if (par2_startnum) {
                        *par2_startnum = atoi(values_par2[1]);
                    }

                    if (num_par2_blocks) {
                        *num_par2_blocks = atoi(values_par2[2]);
                    }

                    nntpgrab_utils_regex_matches_free(values_par2);
                }
            }
        }
    }

    if (part_num) {
        *part_num = atoi(values[3]);
    }

    if (total_parts) {
        *total_parts = atoi(values[4]);
    }

    nntpgrab_utils_regex_matches_free(values);

    return TRUE;
}

char *
nntpgrab_utils_calculate_file_size(guint64 file_size)
{
    double d_file_size;

    d_file_size = (double) file_size;
    if (d_file_size < 1024) {
        return g_strdup_printf("%.1f bytes", d_file_size);
    }

    d_file_size /= 1024;
    if (d_file_size < 1024) {
        return g_strdup_printf("%.1f KB", d_file_size);
    }

    d_file_size /= 1024;
    if (d_file_size < 1024) {
        return g_strdup_printf("%.1f MB", d_file_size);
    }

    d_file_size /= 1024;
    return g_strdup_printf("%.1f GB", d_file_size);
}


int
nntpgrab_utils_calculate_estimated_time_remaining(int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 file_size)
{
    int bytes_per_sec_avg;
    guint64 ret;

    // calculate the average speed of the last 10 seconds
    bytes_per_sec_avg  = bytes_received1;
    bytes_per_sec_avg += bytes_received2;
    bytes_per_sec_avg += bytes_received3;
    bytes_per_sec_avg += bytes_received4;
    bytes_per_sec_avg += bytes_received5;
    bytes_per_sec_avg += bytes_received6;
    bytes_per_sec_avg += bytes_received7;
    bytes_per_sec_avg += bytes_received8;
    bytes_per_sec_avg += bytes_received9;
    bytes_per_sec_avg += bytes_received10;
    bytes_per_sec_avg /= 10;

    // Is there any data received in the last 10 seconds?
    if (bytes_per_sec_avg == 0) {
        return -1;
    }

    // Don't make an estimation when there hasn't been much data seen in the last 10 seconds
    if (bytes_per_sec_avg < 2048) {
        return -1;
    }

    ret = file_size / bytes_per_sec_avg;

    return (int) ret;
}

char *
nntpgrab_utils_get_readable_time_remaining(int estimated_time_remaining)
{
    int num_days;
    int num_hours;
    int num_minutes;
    int num_seconds;
    int left;
    char *days;
    char *hours;
    char *minutes;
    char *seconds;
    char *res;

    if (estimated_time_remaining <= 0) {
        return g_strdup("");
    }

    num_days = estimated_time_remaining / 86400;
    left = estimated_time_remaining % 86400;

    num_hours = left / 3600;
    left %= 3600;

    num_minutes = left / 60;
    num_seconds = left % 60;

    if (num_days == 0) {
        days = NULL;
    } else if (num_days == 1) {
        days = g_strdup_printf(_("%i day"), num_days);
    } else  {
        days = g_strdup_printf(_("%i days"), num_days);
    }

    if (num_hours == 1) {
        hours = g_strdup(_("1 hour"));
    } else {
        hours = g_strdup_printf(_("%i hours"), num_hours);
    }

    if (num_minutes == 1) {
        minutes = g_strdup(_("1 minute"));
    } else {
        minutes = g_strdup_printf(_("%i minutes"), num_minutes);
    }

    if (num_seconds == 0 || num_days > 0) {
        seconds = NULL;
    } else if (num_seconds == 1) {
        seconds = g_strdup(_("1 second"));
    } else {
        seconds = g_strdup_printf(_("%i seconds"), num_seconds);
    }

    if (days) {
        res = g_strdup_printf("%s %s %s", days, hours, minutes);
    } else {
        if (num_hours == 0) {
            if (num_seconds == 0) {
                res = g_strdup_printf("%s", minutes);
            } else {
                res = g_strdup_printf("%s %s", minutes, seconds);
            }
        } else {
            if (num_seconds == 0) {
                res = g_strdup_printf("%s %s", hours, minutes);
            } else {
                res = g_strdup_printf("%s %s %s", hours, minutes, seconds);
            }
        }
    }

    if (days) {
        g_free(days);
    }

    g_free(hours);
    g_free(minutes);

    if (seconds) {
        g_free(seconds);
    }

    return res;
}

char *
nntpgrab_utils_get_readable_finished_time(int estimated_time_remaining)
{
    char res[256];
    struct tm *now;
    time_t stamp;

    if (estimated_time_remaining <= 0) {
        return g_strdup("");
    }

    stamp = time(NULL) + estimated_time_remaining;

    if ((now = localtime(&stamp)) == NULL) {
        // Date could not be parsed
        return g_strdup("");
    }

    if (strftime(res, sizeof(res), "%c", now) == 0) {
        g_error(__FILE__ ":%i buffer too small", __LINE__);
    }

    return g_strdup(res);
}
